# Python DDE interface by David Naylor, http://code.activestate.com/recipes/577654-dde-client/
# Slightly modified to work with python 3 and streamline the code a bit
# I do not understand how this works, just use it for sending/recieving DDE communication through OPUS

from ctypes import POINTER, WINFUNCTYPE, c_char_p, c_void_p, c_int, c_ulong, c_char_p, byref, windll, create_string_buffer
from ctypes.wintypes import BOOL, DWORD, BYTE, INT, LPCWSTR, UINT, ULONG, HWND, MSG

# DECLARE_HANDLE(name) typedef void *name;
HCONV     = c_void_p  # = DECLARE_HANDLE(HCONV)
HDDEDATA  = c_void_p  # = DECLARE_HANDLE(HDDEDATA)
HSZ       = c_void_p  # = DECLARE_HANDLE(HSZ)
LPBYTE    = c_char_p  # POINTER(BYTE)
LPDWORD   = POINTER(DWORD)
LPSTR    = c_char_p
ULONG_PTR = c_ulong

# See windows/ddeml.h for declaration of struct CONVCONTEXT
PCONVCONTEXT = c_void_p

DMLERR_NO_ERROR = 0

# Predefined Clipboard Formats
CF_TEXT         =  1
CF_BITMAP       =  2
CF_METAFILEPICT =  3
CF_SYLK         =  4
CF_DIF          =  5
CF_TIFF         =  6
CF_OEMTEXT      =  7
CF_DIB          =  8
CF_PALETTE      =  9
CF_PENDATA      = 10
CF_RIFF         = 11
CF_WAVE         = 12
CF_UNICODETEXT  = 13
CF_ENHMETAFILE  = 14
CF_HDROP        = 15
CF_LOCALE       = 16
CF_DIBV5        = 17
CF_MAX          = 18

DDE_FACK          = 0x8000
DDE_FBUSY         = 0x4000
DDE_FDEFERUPD     = 0x4000
DDE_FACKREQ       = 0x8000
DDE_FRELEASE      = 0x2000
DDE_FREQUESTED    = 0x1000
DDE_FAPPSTATUS    = 0x00FF
DDE_FNOTPROCESSED = 0x0000

DDE_FACKRESERVED  = (~(DDE_FACK | DDE_FBUSY | DDE_FAPPSTATUS))
DDE_FADVRESERVED  = (~(DDE_FACKREQ | DDE_FDEFERUPD))
DDE_FDATRESERVED  = (~(DDE_FACKREQ | DDE_FRELEASE | DDE_FREQUESTED))
DDE_FPOKRESERVED  = (~(DDE_FRELEASE))

XTYPF_NOBLOCK        = 0x0002
XTYPF_NODATA         = 0x0004
XTYPF_ACKREQ         = 0x0008

XCLASS_MASK          = 0xFC00
XCLASS_BOOL          = 0x1000
XCLASS_DATA          = 0x2000
XCLASS_FLAGS         = 0x4000
XCLASS_NOTIFICATION  = 0x8000

XTYP_ERROR           = (0x0000 | XCLASS_NOTIFICATION | XTYPF_NOBLOCK)
XTYP_ADVDATA         = (0x0010 | XCLASS_FLAGS)
XTYP_ADVREQ          = (0x0020 | XCLASS_DATA | XTYPF_NOBLOCK)
XTYP_ADVSTART        = (0x0030 | XCLASS_BOOL)
XTYP_ADVSTOP         = (0x0040 | XCLASS_NOTIFICATION)
XTYP_EXECUTE         = (0x0050 | XCLASS_FLAGS)
XTYP_CONNECT         = (0x0060 | XCLASS_BOOL | XTYPF_NOBLOCK)
XTYP_CONNECT_CONFIRM = (0x0070 | XCLASS_NOTIFICATION | XTYPF_NOBLOCK)
XTYP_XACT_COMPLETE   = (0x0080 | XCLASS_NOTIFICATION )
XTYP_POKE            = (0x0090 | XCLASS_FLAGS)
XTYP_REGISTER        = (0x00A0 | XCLASS_NOTIFICATION | XTYPF_NOBLOCK )
XTYP_REQUEST         = (0x00B0 | XCLASS_DATA )
XTYP_DISCONNECT      = (0x00C0 | XCLASS_NOTIFICATION | XTYPF_NOBLOCK )
XTYP_UNREGISTER      = (0x00D0 | XCLASS_NOTIFICATION | XTYPF_NOBLOCK )
XTYP_WILDCONNECT     = (0x00E0 | XCLASS_DATA | XTYPF_NOBLOCK)
XTYP_MONITOR         = (0x00F0 | XCLASS_NOTIFICATION | XTYPF_NOBLOCK)

XTYP_MASK            = 0x00F0
XTYP_SHIFT           = 4

TIMEOUT_ASYNC        = 0xFFFFFFFF

def get_winfunc(libname, funcname, restype=None, argtypes=(), _libcache={}):
    """Retrieve a function from a library, and set the data types."""
    if libname not in _libcache:
        _libcache[libname] = windll.LoadLibrary(libname)
    func = getattr(_libcache[libname], funcname)
    func.argtypes = argtypes
    func.restype = restype

    return func


DDECALLBACK = WINFUNCTYPE(HDDEDATA, UINT, UINT, HCONV, HSZ, HSZ, HDDEDATA, 
                          ULONG_PTR, ULONG_PTR)

class DDE(object):
    """Object containing all the DDE functions"""
    AccessData         = get_winfunc("user32", "DdeAccessData",          LPBYTE,   (HDDEDATA, LPDWORD))
    ClientTransaction  = get_winfunc("user32", "DdeClientTransaction",   HDDEDATA, (LPBYTE, DWORD, HCONV, HSZ, UINT, UINT, DWORD, LPDWORD))
    Connect            = get_winfunc("user32", "DdeConnect",             HCONV,    (DWORD, HSZ, HSZ, PCONVCONTEXT))
    CreateStringHandle = get_winfunc("user32", "DdeCreateStringHandleW", HSZ,      (DWORD, LPCWSTR, UINT))
    Disconnect         = get_winfunc("user32", "DdeDisconnect",          BOOL,     (HCONV,))
    GetLastError       = get_winfunc("user32", "DdeGetLastError",        UINT,     (DWORD,))
    Initialize         = get_winfunc("user32", "DdeInitializeW",         UINT,     (LPDWORD, DDECALLBACK, DWORD, DWORD))
    FreeDataHandle     = get_winfunc("user32", "DdeFreeDataHandle",      BOOL,     (HDDEDATA,))
    FreeStringHandle   = get_winfunc("user32", "DdeFreeStringHandle",    BOOL,     (DWORD, HSZ))
    QueryString        = get_winfunc("user32", "DdeQueryStringA",        DWORD,    (DWORD, HSZ, LPSTR, DWORD, c_int))
    UnaccessData       = get_winfunc("user32", "DdeUnaccessData",        BOOL,     (HDDEDATA,))
    Uninitialize       = get_winfunc("user32", "DdeUninitialize",        BOOL,     (DWORD,))

class DDEError(RuntimeError):
    """Exception raise when a DDE errpr occures."""
    def __init__(self, msg, idInst=None):
        if idInst is None:
            RuntimeError.__init__(self, msg)
        else:
            RuntimeError.__init__(self, "%s (err=%s)" % (msg, hex(DDE.GetLastError(idInst))))

class DDEClient(object):
    """The DDEClient class.

    Use this class to create and manage a connection to a service/topic.  To get
    classbacks subclass DDEClient and overwrite callback."""

    def __init__(self, service, topic):
        """Create a connection to a service/topic."""
        self._idInst = DWORD(0)
        self._hConv = HCONV()

        self._callback = DDECALLBACK(self._callback)
        res = DDE.Initialize(byref(self._idInst), self._callback, 0x00000010, 0)
        if res != DMLERR_NO_ERROR:
            raise DDEError("Unable to register with DDEML (err=%s)" % hex(res))

        hszService = DDE.CreateStringHandle(self._idInst, service, 1200)
        hszTopic = DDE.CreateStringHandle(self._idInst, topic, 1200)
        self._hConv = DDE.Connect(self._idInst, hszService, hszTopic, PCONVCONTEXT())
        DDE.FreeStringHandle(self._idInst, hszTopic)
        DDE.FreeStringHandle(self._idInst, hszService)
        if not self._hConv:
            raise DDEError("Unable to establish a conversation with server", self._idInst)

    def __del__(self):
        """Cleanup any active connections."""
        if self._hConv:
            DDE.Disconnect(self._hConv)
        if self._idInst:
            DDE.Uninitialize(self._idInst)

    def advise(self, item, stop=False):
        """Request updates when DDE data changes."""
        hszItem = DDE.CreateStringHandle(self._idInst, item, 1200)
        hDdeData = DDE.ClientTransaction(LPBYTE(), 0, self._hConv, hszItem, CF_TEXT, XTYP_ADVSTOP if stop else XTYP_ADVSTART, TIMEOUT_ASYNC, LPDWORD())
        DDE.FreeStringHandle(self._idInst, hszItem)
        if not hDdeData:
            raise DDEError("Unable to %s advise" % ("stop" if stop else "start"), self._idInst)
        DDE.FreeDataHandle(hDdeData)

    def execute(self, command, timeout=5000):
        """Execute a DDE command."""
        pData = c_char_p(command)
        cbData = DWORD(len(command) + 1)
        hDdeData = DDE.ClientTransaction(pData, cbData, self._hConv, HSZ(), CF_TEXT, XTYP_EXECUTE, timeout, LPDWORD())
        if not hDdeData:
            raise DDEError("Unable to send command", self._idInst)
        DDE.FreeDataHandle(hDdeData)

    def request(self, item, timeout=5000):
        """Request data from DDE service."""
        hszItem = DDE.CreateStringHandle(self._idInst, item, 1200)
        hDdeData = DDE.ClientTransaction(LPBYTE(), 0, self._hConv, hszItem, CF_TEXT, XTYP_REQUEST, timeout, LPDWORD())
        DDE.FreeStringHandle(self._idInst, hszItem)
        if not hDdeData:
            raise DDEError("Unable to request item", self._idInst)

        if timeout != TIMEOUT_ASYNC:
            pdwSize = DWORD(0)
        pData = DDE.AccessData(hDdeData, byref(pdwSize))
        if not pData:
            DDE.FreeDataHandle(hDdeData)
            raise DDEError("Unable to access data", self._idInst)
            # TODO: use pdwSize
            DDE.UnaccessData(hDdeData)
        DDE.FreeDataHandle(hDdeData)
        return pData

    def callback(self, value, item=None):
        """Calback function for advice."""
        print("%s: %s" % (item, value))

    def _callback(self, wType, uFmt, hConv, hsz1, hsz2, hDdeData, dwData1, dwData2):
        dwSize = DWORD(0)
        pData = DDE.AccessData(hDdeData, byref(dwSize))
        if pData:
            item = create_string_buffer('\000' * 128)
            DDE.QueryString(self._idInst, hsz2, item, 128, 1004)
            self.callback(pData, item.value)
            DDE.UnaccessData(hDdeData)
        return DDE_FACK
        return 0

def WinMSGLoop():
    """Run the main windows message loop."""
    LPMSG = POINTER(MSG)
    LRESULT = c_ulong
    GetMessage = get_winfunc("user32", "GetMessageW", BOOL, (LPMSG, HWND, UINT, UINT))
    TranslateMessage = get_winfunc("user32", "TranslateMessage", BOOL, (LPMSG,))
    # restype = LRESULT
    DispatchMessage = get_winfunc("user32", "DispatchMessageW", LRESULT, (LPMSG,))

    msg = MSG()
    lpmsg = byref(msg)
    while GetMessage(lpmsg, HWND(), 0, 0) > 0:
        TranslateMessage(lpmsg)
        DispatchMessage(lpmsg)

if __name__ == "__main__":
    # ------------------------------------------
    # Test with Excel
    # open workbook and enter something in Sheet 1, R1C1
    dde = DDEClient("Excel", "Sheet1")
    print(dde.request("R1C1").decode('ASCII'))
    
    # ------------------------------------------
    # Test with OPUS
    # make sure OPUS software is open
    dde = DDEClient("Opus", "System")
    print(dde.request("GET_VERSION_EXTENDED"))




import os, struct, errno
import numpy as np
from brukeropus.file.constants import STRUCT_3D_INFO_BLOCK


__docformat__ = "google"


def read_opus_file_bytes(filepath):
    '''Returns `bytes` of an OPUS file specified by `filepath` (or `None`).

    Function determines if `filepath` points to an OPUS file by reading the first four bytes which are always the same
    for OPUS files.  If `filepath` is not a file, or points to a non-OPUS file, the function returns `None`.  Otherwise
    the function returns the entire file as raw `bytes`.

    Args:
        filepath (str or Path): full filepath to OPUS file

    Returns:
        **filebytes (bytes):** raw bytes of OPUS file or `None` (if filepath does not point to an OPUS file)
    '''
    filebytes = None
    if os.path.isfile(filepath):
        with open(filepath, 'rb') as f:
            first_four = f.read(4)
            if first_four == b'\n\n\xfe\xfe':
                filebytes = first_four + f.read()
    else:
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), filepath)
    return filebytes


def get_block_type(type_int: int):
    '''Converts an int32 block type code to a six-integer tuple `block_type`.

    This function is used to decode the `type_int` from the directory block of an OPUS file into a tuple of integers.
    Each integer in the tuple provides information about the associated data block.

    Args:
        type_int: 32-bit integer decoded from file directory block

    Returns:
        **block_type (tuple):** six-integer tuple which specifies the block type
    '''
    type_bit_str = format(type_int, '#034b')  # binary representation as string
    block_type = (
        int(type_bit_str[-2:], 2),
        int(type_bit_str[-4:-2], 2),
        int(type_bit_str[-10:-4], 2),
        int(type_bit_str[-17:-10], 2),
        int(type_bit_str[-19:-17], 2),
        int(type_bit_str[-22:-19], 2)
    )
    return block_type


def parse_header(filebytes: bytes):
    '''Parses the OPUS file header.

    The header of an OPUS file contains some basic information about the file including the version number, location of
    the directory block, and number of blocks in the file. This header is first to be parsed as it specifies how to
    read the file directory block (which contains information about each block in the file)

    Args:
        filebytes: raw bytes of OPUS file (all bytes)

    Returns:
        **header_info (tuple):**  
            (  
                **version (float64):** program version number as a floating-point date (later versions always greater)  
                **directory_start (int32):** pointer to start location of directory block (number of bytes)  
                **max_blocks (int32):** maximum number of blocks supported by the directory block (this should only be
                    relevant when trying to edit an OPUS file, i.e. when adding data blocks to a file)  
                **num_blocks (int32):** total number of blocks in the opus file  
            )
    '''
    version = struct.unpack_from('d', filebytes, 4)[0]
    directory_start = struct.unpack_from('<i', filebytes, 12)[0]
    max_blocks = struct.unpack_from('<i', filebytes, 16)[0]
    num_blocks = struct.unpack_from('<i', filebytes, 20)[0]
    return version, directory_start, max_blocks, num_blocks


def parse_directory(filebytes: bytes, directory_start: int, num_blocks: int):
    '''Parses directory block of OPUS file and yields block info for all blocks in the file as a generator.

    The directory block of an OPUS file contains information about every block in the file. The block information is
    stored as three int32 values: `type_int`, `size_int`, `start`.  `type_int` is an integer representation of the block
    type. The bits of this `type_int` have meaning and are parsed into a tuple using `get_block_type`. The `size_int` is
    the size of the block in 32-bit words. `start` is the starting location of the block (in number of bytes).

    Args:
        filebytes: raw bytes of OPUS file (all bytes)
        start: start location of the directory block (specified in file header)
        num_blocks: total number of blocks in the file (specified in file header)

    Yields:
        **block_info (tuple):**  
            (  
                **block_type (tuple):** six-integer tuple which specifies the block type (see: `get_block_type`)  
                **size (int):** size (number of bytes) of the block  
                **start (int):** pointer to start location of the block (number of bytes)  
            )
    '''
    for i in range(num_blocks):
        type_int, size_int, start = struct.unpack_from('<3i', filebytes, directory_start + i*12)
        block_type = get_block_type(type_int)
        size = size_int*4
        yield block_type, size, start


def parse_param_block(filebytes: bytes, size: int, start: int):
    '''Parses the bytes in a parameter block and yields the key, value pairs as a generator.

    Parameter blocks are in the form: `XXX`, `dtype_code`, `size`, `val`.  `XXX` is a three char abbreviation of the
    parameter (key). The value of the parameter is decoded according to the `dtype_code` and size integers to be either:
    `int`, `float`, or `string`.

    Args:
        filebytes: raw bytes of OPUS file (all bytes)
        size: total number of bytes in parameter block (specified in file directory)
        start: pointer to start location of parameter block (specified in file directory)

    Yields:
        **items (tuple):** (key, value) pairs where key is three char string (lowercase) and value can be `int`, `float`
            or `string`.
    '''
    blockbytes = filebytes[start:start + size]
    loc = 0
    while loc < size:
        key = blockbytes[loc:loc+3].decode('utf-8')
        if key == 'END':
            break
        dtype_code, val_size = struct.unpack_from('<2h', blockbytes[loc + 4:loc + 8])
        val_size = val_size * 2
        if dtype_code == 0:
            fmt_str = '<i'
        elif dtype_code == 1:
            fmt_str = '<d'
        else:
            fmt_str = '<'+str(val_size)+'s'
        try:
            val = struct.unpack_from(fmt_str, blockbytes, loc + 8)[0]
            if 's' in fmt_str:
                x00_pos = val.find(b'\x00')
                if x00_pos != -1:
                    val = val[:x00_pos].decode('latin-1')
                else:
                    val = val.decode('latin-1')
        except Exception as e:
            val = 'Failed to load: ' + str(e)
        yield key, val
        loc = loc + val_size + 8


def get_dpf_dtype_count(dpf: int, size: int):
    '''Returns numpy dtype and array count from the data point format (dpf) and block size (in bytes).

    Args:
        dpf: data point format integer stored in data status block.
            dpf = 1 -> array of float32
            dpf = 2 -> array of int32
        size: Block size in bytes.

    Returns:
        **dtype (numpy.dtype):** `numpy` dtype for defining an `ndarray` to store the data
        **count (int):** length of array calculated from the block size and byte size of the dtype.
    '''
    if dpf == 2:
        dtype = np.int32
        count = round(size/4)
    else:
        dtype = np.float32
        count = round(size/4)
    return dtype, count


def parse_data_block(filebytes: bytes, size: int, start: int, dpf=1):
    '''Parses the bytes in a data block (specified by `size` and `start` pointers) and returns a `numpy` array.

    Data blocks contain no metadata, only the y-values of a data array. Data arrays include: single-channel sample,
    reference, phase, interferograms, and a variety of resultant data (transmission, absorption, etc.).  Every data
    block should have a corresponding data status parameter block which can be used to generate the x-array values for
    the data block. The data status block also specifies the data type of the data array with the `DPF` parameter. It
    appears that OPUS currently exclusively stores data blocks as 32-bit floats, but has a reservation for 32-bit
    integers when `DPF` = 2.

    Args:
        filebytes: full OPUS file bytes
        size: size of data block to decode in bytes
        start: pointer to start location of the data block
        dpf: data-point-format integer stored in corresponding data status block.

    Returns:
        **y_array (numpy.ndarray):** `numpy` array of y values contained in the data block
    '''
    dtype, count = get_dpf_dtype_count(dpf=dpf, size=size)
    return np.frombuffer(filebytes[start:], dtype=dtype, count=count)


def parse_3d_data_block(filebytes: bytes, start: int, dpf: int = 1):
    '''Parses the bytes in a 3D data block (series of spectra) and returns a data `dict` containing data and metadata.

    3D data blocks are structured differently than standard data blocks. In addition to the series of spectra, they
    include metadata for each of the spectrum.  This function returns a `dict` containing all the extracted information
    from the data block.  The series spectra is formed into a 2D array while metadata captured for each spectra is
    formed into a 1D array (length = number of spectral measurements in the series).

    Args:
        filebytes: full OPUS file bytes
        start: pointer to start location of the data block
        dpf: data-point-format integer stored in corresponding data status block.

    Returns:
        **data_dict (dict):** `dict` containing all extracted information from the data block  
            {  
                **version:** file format version number (should be 0)  
                **num_blocks:** number of sub blocks; each sub block features a data spectra and associated metadata  
                **offset:** offset in bytes to the first sub data block  
                **data_size:** size in bytes of each sub data block  
                **info_size:** size in bytes of the metadata info block immediately following the sub data block  
                **store_table:** run numbers of the first and last blocks to keep track of skipped spectra  
                **y:** 2D `numpy` array containing all spectra (C-order)  
                **metadata arrays:** series of metadata arrays in 1D array format (e.g. `npt`, `mny`, `mxy`, `tim`).
                    The most useful one is generally `tim`, which can be used as the time axis for 3D data plots.  
            }
    '''
    header = struct.unpack_from('6l', filebytes, start)
    data = {
        'version': header[0],
        'num_blocks': header[1],
        'offset': header[2],
        'data_size': header[3],
        'info_size': header[4],
    }
    data['store_table'] = [struct.unpack_from('<2l', filebytes, start + 24 + i*8) for i in range(header[5])]
    dtype, count = get_dpf_dtype_count(dpf, data['data_size'])
    data['y'] = np.zeros((data['num_blocks'], count), dtype=dtype)
    for entry in STRUCT_3D_INFO_BLOCK:
        data[entry['key']] = np.zeros((data['num_blocks']), dtype=entry['dtype'])
    offset = start + data['offset']
    for i in range(data['num_blocks']):
        data['y'][i] = np.frombuffer(filebytes[offset:], dtype=dtype, count=count)
        offset = offset + data['data_size']
        info_vals = struct.unpack_from('<' + ''.join([e['fmt'] for e in STRUCT_3D_INFO_BLOCK]), filebytes, offset)
        for j, entry in enumerate(STRUCT_3D_INFO_BLOCK):
            data[entry['key']][i] = info_vals[j]
        offset = offset + data['info_size']
    return data


def parse_file_log(filebytes: bytes, size: int, start: int):
    '''Parses the file log in an OPUS file and returns a list of strings contained in the log.

    The file log block of an OPUS file contains some information about how the file was generated and edits that have
    been performed on the file.  This function parses the file log as a list of strings using b'\x00' as a seperator,
    and does not take any steps to parameterizing what is contained in the log.  This log is generally not needed to
    retrieve the file data and metadata, but might be useful for inspecting the file.

    Args:
        filebytes: full OPUS file bytes
        size: size of file log block to decode in bytes
        start: pointer to start location of the file log block

    Returns:
        **strings (list):** list of strings found in the file log.
    '''
    byte_string = struct.unpack_from('<' + str(size) + 's', filebytes, start)[0]
    byte_strings = byte_string.split(b'\x00')
    strings = []
    for entry in byte_strings:
        if entry != b'':
            try:
                strings.append(entry.decode('latin-1'))
            except Exception:
                try:
                    strings.append(entry.decode('utf-8'))
                except Exception as e:
                    strings.append('<Decode Exception>: ' + str(e))
    return strings



'''
`brukeropus` is a Python package for interacting with Bruker's OPUS spectroscopy software. Currently, the package can
read OPUS data files and communicate/control OPUS software using the DDE communication protocol)

### Installation
`brukeropus` requires `python 3.6+` and `numpy`, but `matplotlib` is needed to run the plotting examples.  You can
install with pip:
```python
pip install brukeropus
```

### Namespace
`brukeropus` provides direct imports to the following:
```python
from brukeropus import find_opus_files, read_opus, OPUSFile, Opus
```
All other file functions or classes can be directly imported from the `brukeropus.file` or `brukeropus.control`
submodules, e.g.:
```python
from brukeropus.file import parse_file_and_print
```
It is recommended that you do **not** import from the fully qualified namespace, e.g.:
```python
from brukeropus.file.utils import parse_file_and_print
```
as that namespace is subject to change. Instead import directly from `brukeropus` or its first level submodules.

### Reading OPUS Files (Basic Usage)
```python
from brukeropus import read_opus
from matplotlib import pyplot as plt

opus_file = read_opus('file.0')  # Returns an OPUSFile class

opus_file.print_parameters()  # Pretty prints all metadata in the file to the console

if 'a' in opus_file.data_keys:  # If absorbance spectra was extracted from file
    plt.plot(opus_file.a.x, opus_file.a.y)  # Plot absorbance spectra
    plt.title(opus_file.sfm + ' - ' + opus_file.snm)  # Sets plot title to Sample Form - Sample Name
    plt.show()  # Display plot
```
More detailed documentation on the file submodule can be found in `brukeropus.file`

### Controlling OPUS Software (Basic Usage)
```python
from brukeropus import opus, read_opus
from matplotlib import pyplot as plt

opus = Opus()  # Connects to actively running OPUS software

apt_options = opus.get_param_options('apt') # Get all valid aperture settings

for apt in apt_options[2:-2]: # Loop over all but the two smallest and two largest aperature settings
    filepath = opus.measure_sample(apt=apt, nss=10, unload=True) # Perform measurement and unload file from OPUS
    data = read_opus(filepath) # Read OPUS file from measurement
    plt.plot(data.sm.x, data.sm.y, label=apt) # Plot single-channel sample spectra
plt.legend()
plt.show()
```
More detailed documentation on the control submodule can be found in `brukeropus.control`.
'''

from brukeropus.file import OPUSFile, read_opus, find_opus_files
from brukeropus.control import Opus



import os
from brukeropus.control import DDEClient
from brukeropus import read_opus

__docformat__ = "google"

ERROR_CODES = {
    1: 'Not an Opus Command',
    2: 'Unknown Opus Command',
    3: 'Missing Square Bracket in Command',
    4: 'Function Not Available (Possible missing parameter)',
    5: 'Parameter Name Is Incorrect',
    6: 'Parameter Set Is Incomplete',
    7: 'File Parameter Is Incorrectly Formatted',
    8: 'File(s) Missing Or Corrupt',
    9: 'Opus Could Not Complete The Command',
}


class Opus:
    '''Class for communicating with currently running OPUS software using DDE interface.  Class automatically attempts
    to connect to OPUS software upon initialization.'''
    dde = None
    connected = False
    error_string = 'Error'

    def connect(self):
        '''Connects class to OPUS software through the DDE interface.  Sets the `connected` attribute to `True` if
        successful.  By default, initializing an `Opus` class will automatically attempt to connect to OPUS.'''
        try:
            self.dde = DDEClient("Opus", "System")
            self.connected = True
        except Exception as e:
            self.connected = False
            raise Exception("Failed to connect to OPUS Software: " + str(e))

    def disconnect(self):
        '''Disconnects DDE client/server connection.'''
        if self.connected:
            self.dde.__del__()
            self.dde = None
            self.connected = False

    def raw_query(self, req_str: str, timeout=10000):
        '''Sends command/request string (`req_str`) to OPUS and returns the response in byte format.

        Args:
            req_str: The request string to send to OPUS over DDE
            timeout: timeout in milliseconds.  If a response is not recieved within the timeout period, an exception is
                raised.

        Returns:
            response: response from OPUS software through DDE request in bytes format.'''
        return self.dde.request(req_str, timeout)

    def parse_response(self, byte_response: bytes, decode='ascii'):
        '''Parses the byte response from a raw DDE request query.  If an error is detected in the request, an Exception
        is raised.  If successful, a boolean, string or list of strings will be returned as appropriate.

        Args:
            byte_response: response from OPUS software through DDE request in bytes format.
            decode: format used to decode bytes into string (e.g. 'ascii' or 'utf-8')

        Returns:
            response: parsed response from OPUS software (bool, string, or list of strings depending on request)'''
        str_response = byte_response.decode(decode)
        responses = str_response.split('\n')
        responses = [r for r in responses if r != '']
        if len(responses) == 0:
            raise Exception('Error with DDE request: "' + req_str + '"; no response recieved...')
        elif responses[0].startswith(self.error_string):
            error = self._parse_error(responses[0])
            raise Exception('Error with DDE request: "' + req_str + '"; ' + error)
        else:
            responses = [r for r in responses if r != 'OK']
            if len(responses) == 0:
                return True
            elif len(responses) == 1:
                return responses[0]
            else:
                return responses

    def _parse_error(self, response: str):
        try:
            code = int(response[response.rfind('ID: ') + 4:])
            if code in ERROR_CODES.keys():
                return ERROR_CODES[code]
            else:
                return response
        except:
            return response

    def query(self, req_str: str, timeout=10000, decode='ascii'):
        '''Sends a command/request and returns the parsed response.

        Args:
            req_str: The request string to send to OPUS over DDE
            timeout: timeout in milliseconds.  If a response is not recieved within the timeout period, an exception is
                raised.
            decode: format used to decode bytes into string (e.g. 'ascii' or 'utf-8')

        Returns:
            response: parsed response from OPUS software (bool, string, or list of strings depending on request)
        '''
        response = self.raw_query(req_str=req_str, timeout=timeout)
        return self.parse_response(response, decode=decode)

    def close_opus(self):
        '''Closes the OPUS application. Returns `True` if successful.'''
        return self.query('CLOSE_OPUS')

    def get_param_label(self, param: str):
        '''Get the label for a three character parameter code (e.g. BMS, APT, DTC, etc...).

        Args:
            param: three character parameter code (case insensitive)

        Returns:
            label: short descriptive label that defines the parameter'''
        return self.query('PARAM_STRING ' + param.upper())

    def get_param_options(self, param: str):
        '''Get the parameter setting options for a three character parameter code. Only valid for
        enum type parameters (e.g. BMS, APT, DTC, etc...).

        Args:
            param: three character parameter code (case insensitive)

        Returns:
            options: list of valid options (strings) for the given parameter'''
        result = self.query('ENUM_STRINGS ' + param.upper())
        if type(result) is list:
            return result[1:]
        else:
            return False

    def get_version(self):
        '''Get the OPUS software version information'''
        return self.query('GET_VERSION_EXTENDED')

    def get_opus_path(self):
        '''Get the absolute path to the OPUS software directory (where PARAMTEXT.bin and other instrument specific files
        are located)'''
        return self.query('GET_OPUSPATH')

    def send_command(self, text_command: str, timeout=10000):
        '''Used to send "Direct Commands" to the optics bench. Useful for manually moving motors, etc. from accessories
        and other low-level operations such as controlling the scanning mirror movement.

        Examples:
            send_command('VAC=5') # vents the sample compartment
            send_command('VAC=4') # evacuates sample compartment

        Args:
            text_command: string command as you would enter into "Direct Command" input of OPUS
            timeout: timeout in milliseconds to wait for response

        Returns:
            response: parsed response from OPUS software (typically boolean confirmation)'''

        return self.query('SendCommand(0, {UNI=' + text_command + '})')

    def evacuate_sample(self):
        '''Evacuates the sample compartment'''
        return self.send_command('VAC=4')

    def vent_sample(self):
        '''Vents the sample compartment'''
        return self.send_command('VAC=5')

    def close_flaps(self):
        '''Closes vacumm flaps between optics bench and sample compartment'''
        return self.send_command('FLP=1')

    def open_flaps(self):
        '''Opens vacumm flaps between optics bench and sample compartment'''
        return self.send_command('FLP=0')

    def unload_file(self, filepath: str):
        '''Unloads a file from the OPUS software from its `filepath`

        Args:
            filepath: full path of the file to be unloaded in the software.

        Returns:
            response: `True` if successful.'''
        return self.query('UNLOAD_FILE "' + filepath + '"')

    def unload_all(self):
        '''Unloads all files from OPUS software'''
        return self.query('UnloadAll()')

    def measure_ref(self, timeout=1000000, **kwargs):
        '''Takes a reference measurement using the current settings from advanced experiment.  Also
        takes option **kwargs input which use the OPUS 3-letter parameter keys and values as input
        to customize the measurement.  example:

            measure_ref(nrs=100, res=4) # measures reference with current settings but overriding averages to 100 and
                resolution to 4

        Args:
            timeout: timeout in milliseconds to wait for response
            kwargs: any valid three character parameter code (case insensitive)

        Returns:
            response: `True` if successful
            '''
        params = self._param_str(**kwargs)
        ok = self.query('MeasureReference(0,' + params + ')', timeout=timeout)
        return ok

    def measure_sample(self, unload=False, timeout=1000000, **kwargs):
        '''Takes a reference measurement using the current settings from advanced experiment.  Also
        takes option **kwargs input which use the OPUS 3-letter parameter keys and values as input
        to customize the measurement.  example:

            measure_sample(nss=100, res=4) # measures sample with current settings but overriding averages to 100 and
                resolution to 4

        Args:
            unload: whether to unload the file from OPUS after measurement is complete (to allow moving/renaming, etc.)
            timeout: timeout in milliseconds to wait for response
            kwargs: any valid three character parameter code (case insensitive)

        Returns:
            filepath: absolute filepath to measured sample file'''
        params = self._param_str(**kwargs)
        output = self.query('MeasureSample(0,' + params + ')', timeout=timeout)
        filepath = output[1][1:-3]
        if unload:
            self.unload_file(filepath)
        return filepath

    def check_signal(self, nss=1, **kwargs):
        '''Performs a quick (typically 1 sample) measurement using the current FTIR settings. Current settings can be
        overridden using **kwargs. After measurement is finished, the file is unloaded from OPUS and deleted. The
        function returns an `OPUSFile` object before it deletes the quick measurement file.

        Args:
            nss: number of sample scans to average (default is 1, i.e. no averaging)
            kwargs: any valid three character parameter code (case insensitive)

        Returns:
            opus_file: `OPUSFile` object generated by quick measurement'''
        filepath = self.measure_sample(unload=True, nss=nss, **kwargs)
        opus_file = read_opus(filepath)
        os.remove(filepath)
        return opus_file

    def save_ref(self):
        '''Saves current reference to file (according to current filename and path set in advanced experiment) and
        returns the filename.

        Returns:
            filepath: absolute path to saved reference file'''
        output = self.query('SaveReference()')
        filepath = output[1][1:-3]
        return filepath

    def _param_str(self, **kwargs):
        '''Takes in an arbitrary number of: key=val kwargs and returns a param string of the following format:

        _param_str(nss=100, res=4)
        returns: {NSS=100,RES=4}

        These param strings are used by the measure_ref and measure_sample functions to specify
        experimental parameters.'''
        params = []
        for arg, val in kwargs.items():
            params.append(arg.upper() + '=' + str(val))
        if len(params) > 0:
            return '{' + ','.join(params) + '}'
        else:
            return ''

    def __bool__(self):
        return self.connected

    def __init__(self):
        self.connect()
        if self.connected:
            self.opus_path = self.get_opus_path()



import numpy as np
PARAM_LABELS = {
    'ACC': 'Accessory',
    'ABP': 'Absolute Peak Pos in Laser*2',
    'ADC': 'External Analog Signals',
    'ADT': 'Additional Data Treatment',
    'AG2': 'Actual Signal Gain Channel 2',
    'AN1': 'Analog Signal 1',
    'AN2': 'Analog Signal 2',
    'APF': 'Apodization Function',
    'APR': 'ATR Pressure',
    'APT': 'Aperture Setting',
    'AQM': 'Acquisition Mode',
    'ARG': 'Actual Reference Gain',
    'ARS': 'Number of Reference Scans',
    'ASG': 'Actual Signal Gain',
    'ASS': 'Number of Sample Scans',
    'BBW': 'Number of Bad Backward Scans',
    'BFW': 'Number of Bad Forward Scans',
    'BLD': 'Building',
    'BMS': 'Beamsplitter',
    'CAM': 'Coaddition Mode',
    'CFE': 'Low Intensity Power Mode with DTGS',
    'CHN': 'Measurement Channel',
    'CNM': 'Operator Name',
    'COR': 'Correlation Test Mode',
    'CPG': 'Character Encoding Code Page',
    'CPY': 'Company',
    'CRR': 'Correlation Rejection Reason',
    'CSF': 'Y Scaling Factor',
    'DAQ': 'Data Acquisition Status',
    'DAT': 'Date of Measurement',
    'DEL': 'Delay Before Measurement',
    'DLY': 'Stabilization Delay',
    'DPF': 'Data Point Format',
    'DPM': 'Department',
    'DTC': 'Detector',
    'DUR': 'Duration (sec)',
    'DXU': 'X Units',
    'DYU': 'Y Units',
    'EXP': 'Experiment',
    'FOC': 'Focal Length',
    'FXV': 'First X Value',
    'GBW': 'Number of Good Backward Scans',
    'GFW': 'Number of Good Forward Scans',
    'HFF': 'Digital Filter High Folding Limit',
    'HFL': 'High Folding Limit',
    'HFQ': 'End Frequency Limit for File',
    'HFW': 'Wanted High Freq Limit',
    'HPF': 'High Pass Filter',
    'HUM': 'Relative Humidity Interferometer',
    'INS': 'Instrument Type',
    'IST': 'Instrument Status',
    'LCT': 'Location',
    'LFF': 'Digital Filter Low Folding Limit',
    'LFL': 'Low Folding Limit',
    'LFQ': 'Start Frequency Limit for File',
    'LFW': 'Wanted Low Freq Limit',
    'LPF': 'Low Pass Filter',
    'LPV': 'Variable Low Pass Filter (cm-1)',
    'LWN': 'Laser Wavenumber',
    'LXV': 'Last X Value',
    'MNY': 'Y Minimum',
    'MVD': 'Max Velocity Deviation',
    'MXY': 'Y Maximum',
    'NFL': 'Nominal FW Peak Pos in Points',
    'NLA': 'NL Alpha',
    'NLB': 'NL Beta',
    'NLI': 'Nonlinearity Correction',
    'NPT': 'Number of Data Points',
    'NSN': 'Scan Number',
    'NSR': 'Number of Background Scans',
    'NSS': 'Number of Sample Scans',
    'OPF': 'Optical Filter Setting',
    'P2A': 'Peak Amplitude Channel 2',
    'P2K': 'Backward Peak Location Channel 2',
    'P2L': 'Peak Location Channel 2',
    'P2R': 'Backward Peak Amplitude Channel 2',
    'PGN': 'Preamplifier Gain',
    'PGR': 'Reference Preamplifier Gain',
    'PHR': 'Phase Resolution',
    'PHZ': 'Phase Correction Mode',
    'PKA': 'Peak Amplitude',
    'PKL': 'Peak Location',
    'PLF': 'Result Spectrum Type',
    'PRA': 'Backward Peak Amplitude',
    'PRL': 'Backward Peak Location',
    'PRS': 'Pressure Interferometer (hPa)',
    'RCH': 'Reference Measurement Channel',
    'RDX': 'Extended Ready Check',
    'RDY': 'Ready Check',
    'RES': 'Resolution (cm-1)',
    'RG2': 'Signal Gain, Background 2nd Channel',
    'RGN': 'Reference Signal Gain',
    'RSN': 'Running Sample Number',
    'SFM': 'Sample Form',
    'SG2': 'Signal Gain, Sample 2nd Channel',
    'SGN': 'Sample Signal Gain',
    'SNM': 'Sample Name',
    'SON': 'External Sync',
    'SOT': 'Sample Scans or Time',
    'SPO': 'Sample Number',
    'SPZ': 'Stored Phase Mode',
    'SRC': 'Source',
    'SRN': 'Instrument Serial Number',
    'SRT': 'Start Time (sec)',
    'SSM': 'Sample Spacing Multiplier',
    'SSP': 'Sample Spacing Divisor',
    'STR': 'Scans or Time (Reference)',
    'TCL': 'Command Line for Additional Data Treatment',
    'TDL': 'To Do List',
    'TIM': 'Time of Measurement',
    'TPX': 'Total Points X',
    'TSC': 'Scanner Temperature',
    'UID': 'Universally Unique Identifier',
    'VEL': 'Scanner Velocity',
    'VSN': 'Firmware Version',
    'WAS': 'Tr.Rec. Slices',
    'WDV': 'Transient Recorder',
    'WIB': 'Tr.Rec.Input Range 2nd channel',
    'WIR': 'Tr.Rec.Input Range',
    'WPD': 'Tr.Rec. Stab. Delay after Stepping',
    'WRC': 'Tr.Rec. Repeat Count',
    'WSS': 'Tr.Rec. Sampling Source',
    'WTD': 'Tr.Rec. trigger Delay in points',
    'WTR': 'Tr.Rec. Resolution',
    'WXD': 'Tr.Rec. Experiment Delay',
    'WXP': 'Tr.Rec. Trigger Mode',
    'XPP': 'Experiment Path',
    'XSM': 'Xs Sampling Mode',
    'ZFF': 'Zero Filling Factor',
    }


CODE_0 = {0: '',
          1: 'Real Part of Complex Data',
          2: 'Imaginary Part of Complex Data',
          3: '',  # Amplitude - leave blank because it is redundant when forming a label
          }


CODE_1 = {0: '',
          1: 'Sample',
          2: 'Reference',
          3: '',  # Ratioed - leave blank because it is redundant when forming a label
          }


CODE_2 = {0: '',
          1: 'Data Status Parameters',
          2: 'Instrument Status Parameters',
          3: 'Acquisition Parameters',
          4: 'Fourier Transform Parameters',
          5: 'Plot and Display Parameters',
          6: 'Optical Parameters',
          7: 'GC Parameters',
          8: 'Library Search Parameters',
          9: 'Communication Parameters',
          10: 'Sample Origin Parameters',
          }


CODE_3 = {0: '',
          1: 'Spectrum',
          2: 'Interferogram',
          3: 'Phase',
          4: 'Absorbance',
          5: 'Transmittance',
          6: 'Kubelka-Munk',
          7: 'Trace (Intensity over time)',
          8: 'gc File, Series of Interferograms',
          9: 'gc File, Series of Spectra',
          10: 'Raman',
          11: 'Emisson',
          12: 'Reflectance',
          13: 'Directory',
          14: 'Power',
          15: 'log Reflectance',
          16: 'ATR',
          17: 'Photoacoustic',
          18: 'Result of Arithmatics, looks like Transmittance',
          19: 'Result of Arithmatics, looks like Absorbance',
          }


CODE_4 = {0: '',
          1: 'First Derivative',
          2: 'Second Derivative',
          3: 'n-th Derivative',
          }


CODE_5 = {0: '',
          1: 'Compound Information',
          2: 'Peak Table',
          3: 'Molecular Structure',
          4: 'Macro',
          5: 'File Log',
          }

# care must be taken when using 3-letter keys to avoid cross contaminating with 3-char parameter keys
CODE_3_ABR = {0: '',
              1: '',  # sc - single channel  leave blank because it is redundant when paired with sm, rf
              2: 'ig',
              3: 'ph',
              4: 'a',
              5: 't',
              6: 'km',
              7: 'tr',
              8: 'gcig',
              9: 'gcsc',
              10: 'ra',
              11: 'e',
              12: 'r',
              13: 'dir',
              14: 'p',
              15: 'logr',
              16: 'atr',
              17: 'pas',
              18: 'arit',
              19: 'aria',
              }


TYPE_CODE_LABELS = [CODE_0, CODE_1, CODE_2, CODE_3, CODE_4, CODE_5]


STRUCT_3D_INFO_BLOCK = [
    {'key': 'nss', 'fmt': 'l', 'dtype': np.int32},
    {'key': 'nsr', 'fmt': 'l', 'dtype': np.int32},
    {'key': 'nsn', 'fmt': 'l', 'dtype': np.int32},
    {'key': 'npt', 'fmt': 'l', 'dtype': np.int32},
    {'key': 'gfw', 'fmt': 'l', 'dtype': np.int32},
    {'key': 'gbw', 'fmt': 'l', 'dtype': np.int32},
    {'key': 'bfw', 'fmt': 'l', 'dtype': np.int32},
    {'key': 'bbw', 'fmt': 'l', 'dtype': np.int32},
    {'key': 'hfl', 'fmt': 'd', 'dtype': np.float64},
    {'key': 'lfl', 'fmt': 'd', 'dtype': np.float64},
    {'key': 'hff', 'fmt': 'd', 'dtype': np.float64},
    {'key': 'lff', 'fmt': 'd', 'dtype': np.float64},
    {'key': 'filter_size', 'fmt': 'l', 'dtype': np.int32},
    {'key': 'filter_type', 'fmt': 'l', 'dtype': np.int32},
    {'key': 'fxv', 'fmt': 'd', 'dtype': np.float64},
    {'key': 'lxv', 'fmt': 'd', 'dtype': np.float64},
    {'key': 'mny', 'fmt': 'd', 'dtype': np.float64},
    {'key': 'mxy', 'fmt': 'd', 'dtype': np.float64},
    {'key': 'csf', 'fmt': 'd', 'dtype': np.float64},
    {'key': 'pka', 'fmt': 'd', 'dtype': np.float64},
    {'key': 'pra', 'fmt': 'd', 'dtype': np.float64},
    {'key': 'pkl', 'fmt': 'l', 'dtype': np.int32},
    {'key': 'prl', 'fmt': 'l', 'dtype': np.int32},
    {'key': 'srt', 'fmt': 'd', 'dtype': np.float64},
    {'key': 'tim', 'fmt': 'd', 'dtype': np.float64},
]


Y_LABELS = {
    'sm': 'Sample Spectrum',
    'rf': 'Reference Spectrum',
    'igsm': 'Sample Interferogram',
    'igrf': 'Reference Interferogram',
    'phsm': 'Sample Phase',
    'phrf': 'Reference Phase',
    'a': 'Absorbance',
    't': 'Transmittance',
    'r': 'Reflectance',
    'km': 'Kubelka-Munk',
    'tr': 'Trace (Intensity over Time)',
    'gcig': 'gc File (Series of Interferograms)',
    'gcsc': 'gc File (Series of Spectra)',
    'ra': 'Raman',
    'e': 'Emission',
    'dir': 'Directory',
    'p': 'Power',
    'logr': 'log(Reflectance)',
    'atr': 'ATR',
    'pas': 'Photoacoustic',
    }


XUN_LABELS = {
    'wl': 'Wavelength',
    'wn': 'Wavenumber',
    'f': 'Frequency',
    'pnt': 'Points',
    'min': 'Minutes',
    'logwn': 'Log Wavenumber',
    }



'''
The `brukeropus.file` submodule of `brukeropus` includes all the functions and classes for reading and exploring OPUS
files. This includes both high-level functions like `read_opus` that returns an `OPUSFile` class, as well as low-level
parsing functions like `parse_directory` that returns data extracted directly from the binary OPUS file bytes.  This
overview documentation will focus on the high-level functions which will be useful for most users.  If you are
interested in using the low-level parsing functions, perhaps to make your own data class or customize how files are
read, refer to: `brukeropus.file.parser` which contains all the low-level parsing functions.
## Finding OPUS Files
OPUS files are typically saved with a numeric file extension (e.g. file.0, file.1, file.1001).  This makes searching for
a list of OPUS files in a directory a little more cumbersome than a traditional "*.csv" search.  To address this,
`brukeropus` includes a `find_opus_files` function:
```python
from brukeropus import find_opus_files

filepaths = find_opus_files(r'path\\to\\opus\\files', recursive=True)
```
Which will assign a list of filepaths that match the numeric extension formatting of OPUS files. For full documentation,
see `brukeropus.file.utils.find_opus_files`.
## Reading OPUS Files
`brukeropus` parses OPUS files and assembles them into an `OPUSFile` object that contains the extracted data (and
metadata) within the file.  You can generate an `OPUSFile` object in one of two ways:
```python
from brukeropus import read_opus, OPUSFile

filepath = r'path\\to\\opusfile.0'

data = read_opus(filepath)
same_data = OPUSFile(filepath)
```
In the above code, `data` and `same_data` are both `OPUSFile` objects with identical data.
## Using the `OPUSFile` Class
OPUS files all start with the same first four *magic bytes*.  If the file does not start with these bytes (i.e. is not
a valid OPUS file), the OPUSFile class will logically evaluate to `false`:
```python
data = read_opus('file.pdf')
if data:
    print(data)
else:
    print(data.filepath, 'is not an OPUS file')
```
To view all parameter metadata in the file, you can print to the console using the class method: `print_parameters`.
This will let you view all the key, value parameter data extracted from the file with labels for what the parameter keys
are referring to wherever known.

```python
data = read_opus('file.0')
data.print_parameters()
```
<details>
<summary>Example `print_parameters` Output</summary>
<p>
```console
====================================================================================================
                                         Optical Parameters
Key    Label                                   Value
ACC    Accessory                               TRANS *010A984F
APR    ATR Pressure                            0
APT    Aperture Setting                        1 mm
BMS    Beamsplitter                            KBr-Broadband
CHN    Measurement Channel                     Sample Compartment
DTC    Detector                                RT-DLaTGS [Internal Pos.1]
HPF    High Pass Filter                        0
LPF    Low Pass Filter                         10.0
LPV    Variable Low Pass Filter (cm-1)         4000
OPF    Optical Filter Setting                  Open
PGN    Preamplifier Gain                       3
RDX    Extended Ready Check                    0
SRC    Source                                  MIR
VEL    Scanner Velocity                        10.0
ADC    External Analog Signals                 0
SON    External Sync                           Off

====================================================================================================
                                    Fourier Transform Parameters
Key    Label                                   Value
APF    Apodization Function                    B3
HFQ    End Frequency Limit for File            500.0
LFQ    Start Frequency Limit for File          10000.0
NLI    Nonlinearity Correction                 0
PHR    Phase Resolution                        100.0
PHZ    Phase Correction Mode                   ML
SPZ    Stored Phase Mode                       NO
ZFF    Zero Filling Factor                     2

====================================================================================================
                                       Acquisition Parameters
Key    Label                                   Value
ADT    Additional Data Treatment               0
AQM    Acquisition Mode                        DD
CFE    Low Intensity Power Mode with DTGS      0
COR    Correlation Test Mode                   0
DEL    Delay Before Measurement                0
DLY    Stabilization Delay                     0
HFW    Wanted High Freq Limit                  15000.0
LFW    Wanted Low Freq Limit                   0.0
NSS    Number of Sample Scans                  50
PLF    Result Spectrum Type                    AB
RES    Resolution (cm-1)                       4.0
SOT    Sample Scans or Time                    0
TCL    Command Line for Additional Data Tr...
TDL    To Do List                              16777271
SGN    Sample Signal Gain                      1

====================================================================================================
                                      Sample Origin Parameters
Key    Label                                   Value
BLD    Building
CNM    Operator Name                           Duran
CPY    Company
DPM    Department
EXP    Experiment                              MWIR-LWIR_Trans_FileNameFormat.XPM
LCT    Location
SFM    Sample Form                             Atm-MWIR (All A)
SNM    Sample Name                             File Test
XPP    Experiment Path                         C:\\Users\\Public\\Documents\\Bruker\\OPUS_8.1.29\\XPM
IST    Instrument Status                       OK
CPG    Character Encoding Code Page            1252
UID    Universally Unique Identifier           0d1348c2-3a2c-41c9-b521-bdaf0a23710c

====================================================================================================
                                    Instrument Status Parameters
Key    Label                                   Value
HFL    High Folding Limit                      15795.820598
LFL    Low Folding Limit                       0.0
LWN    Laser Wavenumber                        15795.820598
ABP    Absolute Peak Pos in Laser*2            52159
SSP    Sample Spacing Divisor                  1
ASG    Actual Signal Gain                      1
ARG    Actual Reference Gain                   1
ASS    Number of Sample Scans                  50
GFW    Number of Good Forward Scans            25
GBW    Number of Good Backward Scans           25
BFW    Number of Bad Forward Scans             0
BBW    Number of Bad Backward Scans            0
PKA    Peak Amplitude                          1409
PKL    Peak Location                           7364
PRA    Backward Peak Amplitude                 1356
PRL    Backward Peak Location                  7363
P2A    Peak Amplitude Channel 2                1
P2L    Peak Location Channel 2                 1
P2R    Backward Peak Amplitude Channel 2       1
P2K    Backward Peak Location Channel 2        1
DAQ    Data Acquisition Status                 0
AG2    Actual Signal Gain Channel 2            1
HUM    Relative Humidity Interferometer        14
SSM    Sample Spacing Multiplier               1
RSN    Running Sample Number                   565
CRR    Correlation Rejection Reason            0
SRT    Start Time (sec)                        1556890484.642
DUR    Duration (sec)                          42.433990478515625
TSC    Scanner Temperature                     27.8
MVD    Max Velocity Deviation                  0.1158025860786438
PRS    Pressure Interferometer (hPa)           1009.9999700000001
AN1    Analog Signal 1                         0.22596596493037535
AN2    Analog Signal 2                         3.459206583321489
VSN    Firmware Version                        2.450 Oct 10 2014
SRN    Instrument Serial Number                1135
CAM    Coaddition Mode                         0
INS    Instrument Type                         VERTEX 80V
FOC    Focal Length                            100.0
RDY    Ready Check                             1

====================================================================================================
                               Reference Instrument Status Parameters
Key    Label                                   Value
HFL    High Folding Limit                      15795.820598
LFL    Low Folding Limit                       0.0
LWN    Laser Wavenumber                        15795.820598
ABP    Absolute Peak Pos in Laser*2            52159
SSP    Sample Spacing Divisor                  1
ARG    Actual Reference Gain                   1
ASG    Actual Signal Gain                      1
ASS    Number of Sample Scans                  1
GFW    Number of Good Forward Scans            1
GBW    Number of Good Backward Scans           0
BFW    Number of Bad Forward Scans             0
BBW    Number of Bad Backward Scans            0
PKA    Peak Amplitude                          1644
PKL    Peak Location                           7364
PRA    Backward Peak Amplitude                 1
PRL    Backward Peak Location                  -1
P2A    Peak Amplitude Channel 2                1
P2L    Peak Location Channel 2                 1
P2R    Backward Peak Amplitude Channel 2       1
P2K    Backward Peak Location Channel 2        1
DAQ    Data Acquisition Status                 0
AG2    Actual Signal Gain Channel 2            1
HUM    Relative Humidity Interferometer        0
SSM    Sample Spacing Multiplier               1
RSN    Running Sample Number                   5816
CRR    Correlation Rejection Reason            0
SRT    Start Time (sec)                        1556890282.358
DUR    Duration (sec)                          0.7919998168945312
TSC    Scanner Temperature                     27.8
MVD    Max Velocity Deviation                  0.10553144663572311
PRS    Pressure Interferometer (hPa)           2.01999
AN1    Analog Signal 1                         0.22577181458473206
AN2    Analog Signal 2                         4.0960001945495605
VSN    Firmware Version                        2.450 Oct 10 2014
SRN    Instrument Serial Number                1135
CAM    Coaddition Mode                         0
INS    Instrument Type                         VERTEX 80V
FOC    Focal Length                            100.0
RDY    Ready Check                             1
ARS    Number of Reference Scans               1

====================================================================================================
                                    Reference Optical Parameters
Key    Label                                   Value
ACC    Accessory                               TRANS *010A984F
APR    ATR Pressure                            0
APT    Aperture Setting                        1 mm
BMS    Beamsplitter                            KBr-Broadband
DTC    Detector                                RT-DLaTGS [Internal Pos.1]
HPF    High Pass Filter                        0
LPF    Low Pass Filter                         10.0
LPV    Variable Low Pass Filter (cm-1)         4000
OPF    Optical Filter Setting                  Open
PGR    Reference Preamplifier Gain             3
RCH    Reference Measurement Channel           Sample Compartment
RDX    Extended Ready Check                    0
SRC    Source                                  MIR
VEL    Scanner Velocity                        10.0
ADC    External Analog Signals                 0
SON    External Sync                           Off

====================================================================================================
                                  Reference Acquisition Parameters
Key    Label                                   Value
ADT    Additional Data Treatment               0
AQM    Acquisition Mode                        DD
CFE    Low Intensity Power Mode with DTGS      0
COR    Correlation Test Mode                   0
DEL    Delay Before Measurement                0
DLY    Stabilization Delay                     0
HFW    Wanted High Freq Limit                  15000.0
LFW    Wanted Low Freq Limit                   0.0
NSR    Number of Background Scans              1
PLF    Result Spectrum Type                    TR
RES    Resolution (cm-1)                       4.0
RGN    Reference Signal Gain                   1
STR    Scans or Time (Reference)               0
TCL    Command Line for Additional Data Tr...
TDL    To Do List                              16777271

====================================================================================================
                               Reference Fourier Transform Parameters
Key    Label                                   Value
APF    Apodization Function                    B3
HFQ    End Frequency Limit for File            500.0
LFQ    Start Frequency Limit for File          10000.0
NLI    Nonlinearity Correction                 0
PHR    Phase Resolution                        100.0
PHZ    Phase Correction Mode                   ML
SPZ    Stored Phase Mode                       NO
ZFF    Zero Filling Factor                     2
```
</p>
</details>
  
You can access a specific parameter simply by calling the key as a direct attribute of the class (case insensitive). You
can also get the human-readable label using the `get_param_label` function:

```python
from brukeropus.file import get_param_label
data = read_opus('file.0')
print(get_param_label('bms') + ':', data.bms)
print(get_param_label('src') + ':', data.src)
```
```console
Beamsplitter: KBr-Broadband
Source: MIR
```

You will notice in the example output that some keys (e.g. zero filling factor `zff`) may have two entries: one for the
sample measurement and another for the reference.  By default, the sample parameters are accessible directly from the
`OPUSFile` class, while the reference parameters can be accessed through the `rf_params` attribute.

```python
data = read_opus('file.0')
print('Sample ZFF:', data.zff, 'Reference ZFF:', data.rf_params.zff)
```
```console
Sample ZFF: 2 Reference ZFF: 2
```
You can also iterate over the parameters using the familiar `keys()`, `values()`, and `items()` functions using the
`params` or `rf_params` attributes:

```python
data = read_opus('file.0')
for key, val in data.params.items():
    print(key + ':', val)
```
```console
acc: TRANS *010A984F
apr: 0
apt: 1 mm
bms: KBr-Broadband
chn: Sample Compartment
dtc: RT-DLaTGS [Internal Pos.1]
hpf: 0
lpf: 10.0
lpv: 4000
opf: Open
pgn: 3
... continued ...
```

Depending on the settings used to save the OPUS file, different data blocks can be stored.  To retrieve a list of data
blocks stored in the OPUS File, use the `data_keys` attribute:

```python
data = read_opus('file.0')
print(data.data_keys)
```
```console
['igsm', 'phsm', 'sm', 'a', 'igrf', 'rf']
```

Each key is also an attribute of the `OPUSFile` instance that returns either a `Data` or `Data3D` class.  You can get
the `x` and `y` array values (in the units they were saved in) as direct attributes to the `Data` class:

```python
data = read_opus('file.0')
plt.plot(data.a.x, data.a.y)
plt.ylim((0, 1))
plt.show()
```

For spectra with wavenumber as valid unit (e.g. single-channel or ratioed spectra), the `x` array can be given in `cm⁻¹`
or `µm` units by using the attributes `wn` or `wl` respectively:

```python
data = read_opus('file.0')
plt.plot(data.sm.wl, data.sm.y)
plt.show()
```

You can also iterate over all data spectra in the file using the `iter_data()` method:

```python
data = read_opus('file.0')
for d in data.iter_data():
    print(d.label, '(' + d.datetime.isoformat(' ') + ')')
```
```console
Sample Interferogram (2019-05-03 13:34:44.641000)
Sample Phase (2019-05-03 13:34:44.641000)
Sample Spectrum (2019-05-03 13:34:44.641000)
Absorbance (2019-05-03 13:34:44.641000)
Reference Interferogram (2019-05-03 13:31:22.358000)
Reference Spectrum (2019-05-03 13:31:22.358000)
```

Each data block in an OPUS file also contains a small parameter block with information such as the min/max y-value
(mny, mxy), x-units (dxu), number of data points (npt), etc.  These can be accessed as direct attributes to the `Data`
class, or through the `params` attribute:

```python
data = read_opus('file.0')
print('Sample spectra y-min:', data.sm.mny, 'y-max:', data.sm.mxy)
```
```console
Sample spectra y-min: 1.2147593224653974e-05 y-max: 0.03543896973133087
```
For full API documentation, see:  
`OPUSFile`: `brukeropus.file.file.OPUSFile`  
`Data`: `brukeropus.file.file.Data`  
`Data3D`: `brukeropus.file.file.Data3D`
'''
from brukeropus.file.file import *
from brukeropus.file.parser import *
from brukeropus.file.utils import *
from brukeropus.file.constants import *



import datetime, warnings
import numpy as np
from brukeropus.file.utils import get_block_type_label, get_data_key, get_param_label, _print_block_header, _print_cols
from brukeropus.file.parser import (read_opus_file_bytes,
                                    parse_header,
                                    parse_directory,
                                    parse_data_block,
                                    parse_3d_data_block,
                                    parse_param_block,
                                    parse_file_log)


__docformat__ = "google"


def read_opus(filepath):
    '''Return an `OPUSFile` object from an OPUS file filepath.

    The following produces identical results:
        ```python
        data = read_opus(filepath)
        data = OPUSFile(filepath)
        ```
    Args:
        filepath (str or Path): filepath of an OPUS file (typically *.0)

    Returns:
        opus_file (`OPUSFile`): an instance of the `OPUSFile` class containing all data/metadata extracted from the
        file.
    '''
    return OPUSFile(filepath)


class OPUSFile:
    '''Class that contains the data and metadata contained in a bruker OPUS file.

    Args:
        filepath: full path to the OPUS file to be parsed. Can be a string or Path object and is required to initilize
            an `OPUSFile` object.

    Attributes:
        is_opus (`bool`): True if filepath points to an OPUS file, False otherwise. Also returned for dunder 
            `__bool__()`  
        params (`Parameters`): class containing all general parameter metadata for the OPUS file. To save typing, the
            three char parameters from params also become attributes of the `OPUSFile` class (e.g. bms, apt, src)  
        rf_params (`Parameters`): class containing all reference parameter metadata for the OPUS file.  
        data_keys (list): list of all data block keys stored in the file (i.e. sm, rf, t, a, r, igsm, igrf, phsm, etc.).
            These keys become data attributes of the class which return an instance of `Data` or `Data3D`.
        datetime (`datetime`): Returns the most recent datetime of all the data blocks stored in the file (typically
            result spectra)
        directory (`FileDirectory`):  class containing information about all the various data blocks in the file.
        file_log (str): File log containing text about how the file was generated/edited (not always saved)

    Data Attributes:
        **sm:** Single-channel sample spectra  
        **rf:** Single-channel reference spectra  
        **igsm:** Sample interferogram  
        **igrf:** Reference interferogram  
        **phsm:** Sample phase  
        **phrf:** Reference phase  
        **a:** Absorbance  
        **t:** Transmittance  
        **r:** Reflectance  
        **km:** Kubelka-Munk  
        **tr:** Trace (Intensity over Time)  
        **gcig:** gc File (Series of Interferograms)  
        **gcsc:** gc File (Series of Spectra)  
        **ra:** Raman  
        **e:** Emission  
        **dir:** Directory  
        **p:** Power  
        **logr:** log(Reflectance)  
        **atr:** ATR  
        **pas:** Photoacoustic  
    '''

    def __init__(self, filepath: str):
        self.filepath = filepath
        filebytes = read_opus_file_bytes(filepath)
        if filebytes is None:
            self.is_opus = False
        else:
            self.is_opus = True
            self.directory = FileDirectory(filebytes)
            self.params = Parameters(filebytes, self.directory.param_blocks)
            self.rf_params = Parameters(filebytes, self.directory.rf_param_blocks)
            if hasattr(self.directory, 'file_log_block'):
                self.file_log = '\n'.join(parse_file_log(filebytes,
                                                            self.directory.file_log_block.size,
                                                            self.directory.file_log_block.start))
            self.data_keys = []
            for data, status in self.directory.data_and_status_block_pairs:
                key = data.get_data_key()
                if data.is_3d_data():
                    data_class = Data3D(filebytes, data, status)
                else:
                    data_class = Data(filebytes, data, status)
                    if len(data_class.y) < data_class.npt:
                        break  # Don't add data blocks with missing points (rare but observed)
                if 'vel' in self.params.keys():
                    data_class.vel = self.params.vel
                setattr(self, key, data_class)
                self.data_keys.append(key)

    def __str__(self):
        if self.is_opus:
            data_str = ', '.join(self.data_keys)
            return 'OPUS File: spectra: ' + data_str + '   path: ' + str(self.filepath)
        else:
            return 'Not an OPUS file: ' + str(self.filepath)

    def __bool__(self):
        return self.is_opus

    def __getattr__(self, name):
        if name.lower() in self.params.keys():
            return getattr(self.params, name.lower())
        elif name == 'datetime':
            return max(d.datetime for d in self.iter_data() if d.datetime is not None)
        else:
            text = str(name) + ' is not a valid attribute for OPUSFile: ' + str(self.filepath)
            raise AttributeError(text)

    def print_parameters(self, key_width=7, label_width=40, value_width=53):
        '''Prints all the parameter metadata to the console (organized by block)'''
        width = key_width + label_width + value_width
        col_widths = (key_width, label_width, value_width)
        param_blocks = self.directory.param_blocks + self.directory.rf_param_blocks
        for block in param_blocks:
            label = get_block_type_label(block.type)
            _print_block_header(label, width=width)
            _print_cols(('Key', 'Label', 'Value'), col_widths=col_widths)
            for key in block.keys:
                label = get_param_label(key)
                if block.is_rf_param():
                    value = getattr(self.rf_params, key)
                else:
                    value = getattr(self.params, key)
                _print_cols((key.upper(), label, value), col_widths=col_widths)

    def iter_data(self):
        '''Generator that yields the various Data classes from the OPUSFile'''
        for key in self.data_keys:
            yield getattr(self, key)


class FileBlockInfo:
    '''Contains type, size and location information about an OPUS file block.

    This information is parsed from the directory block of an OPUS file and provides the information needed to parse the
    block.

    Args:
        block_type: six integer tuple that describes the type of data in the file block
        size: size of block in number of bytes
        start: pointer to start location of the block within the file.

    Attributes:
        type: six integer tuple that describes the type of data in the file block
        size: size of block in number of bytes
        start: pointer to start location of the block within the file
        keys: tuple of three char keys contained in parameter blocks. This attribute is set by the OPUSFile class only
            when the block is parameter block. This enables grouping parameters by block if desired.
        bytes: raw bytes of file block (currently only set for unknown blocks)
    '''

    __slots__ = ('type', 'size', 'start', 'keys', 'bytes')

    keys: tuple

    def __init__(self, block_type: tuple, size: int, start: int):
        self.type = block_type
        self.size = size
        self.start = start

    def __str__(self):
        label = self.get_label()
        return 'Block Info: ' + label + ' (size: ' + str(self.size) + ' bytes; start: ' + str(self.start) + ')'

    def is_data_status(self):
        '''Returns True if FileBlockInfo is a data status parameter block'''
        return self.type[2] == 1

    def is_rf_param(self):
        '''Returns True if FileBlockInfo is a parameter block associated with the reference measurement'''
        return self.type[2] > 1 and self.type[1] == 2

    def is_param(self):
        '''Returns True if FileBlockInfo is a parameter block'''
        return self.type[2] > 1 or self.type == (0, 0, 0, 0, 0, 1)

    def is_directory(self):
        '''Returns True if FileBlockInfo is the directory block'''
        return self.type == (0, 0, 0, 13, 0, 0)

    def is_file_log(self):
        '''Returns True if FileBlockInfo is the file log block'''
        return self.type == (0, 0, 0, 0, 0, 5)

    def is_data(self):
        '''Returns True if FileBlockInfo is a data block or 3D data block'''
        return self.type[2] == 0 and self.type[3] not in [0, 13]

    def is_3d_data(self):
        '''Returns True if FileBlockInfo is a 3D data block (i.e. data series)'''
        return self.type[2] == 0 and self.type[5] == 2

    def is_data_status_match(self, data_block_info):
        '''Returns True if FileBlockInfo is a data status block and a match to the data_block_info argument.

        This function is used to match a data status block (contains metadata for data block) with its associated data
        block (contains array data).

        Args:
            data_block_info (FileBlockInfo):  data block being tested as a match.

        Returns:
            is_match (bool): True if FileBlockInfo is data status block and input argument is matching data block'''
        if self.is_data_status():
            return data_block_info.type[:2] == self.type[:2] and data_block_info.type[3:] == self.type[3:]

    def get_label(self):
        '''Returns a friendly string label that describes the block type'''
        return get_block_type_label(self.type)

    def get_data_key(self):
        '''If block is a data block, this function will return an shorthand key to reference that data.

        e.g. t: transmission, a: absorption, sm: sample, rf: reference, smph: sample phase etc. If the block is not
        a data block, it will return None.'''
        if self.is_data():
            return get_data_key(self.type)
        else:
            return None


class Data:
    '''Class containing array data and associated parameter/metadata from an OPUS file.

    Args:
        filebytes: raw bytes from OPUS file. see: `read_opus_file_bytes`
        data_info: `FileBlockInfo` instance of a data block
        data_status_info: `FileBlockInfo` instance of a data status block which contains metadata about the data_info
            block. This block is a parameter block.

    Attributes:
        params: `Parameter` class with metadata associated with the data block such as first x point: `fxp`, last x
            point: `lxp`, number of points: `npt`, date: `dat`, time: `tim` etc.
        y: 1D `numpy` array containing y values of data block
        x: 1D `numpy` array containing x values of data block. Units of x array are given by `dxu` parameter.
        label: human-readable string label describing the data block (e.g. Sample Spectrum, Absorbance, etc.)

    Extended Attributes:
        **wn:** Returns the x array in wavenumber (cm⁻¹) units regardless of what units the x array was originally
            saved in. This is only valid for spectral data blocks such as sample, reference, transmission, etc., not
            interferogram or phase blocks.  
        **wl:** Returns the x array in wavelength (µm) units regardless of what units the x array was originally
            saved in. This is only valid for spectral data blocks such as sample, reference, transmission, etc., not
            interferogram or phase blocks.  
        **f:** Returns the x array in modulation frequency units (Hz) regardless of what units the x array was
            originally saved in. This is only valid for spectral data blocks such as sample, reference, transmission,
            etc., not interferogram or phase blocks.  
        **datetime:** Returns a `datetime` class of when the data was taken (extracted from data status parameter block).  
        **xxx:** the various three char parameter keys from the `params` attribute can be directly called from the 
            `Data` class for convenience. Common parameters include `dxu` (x units), `mxy` (max y value), `mny` (min y
            value), etc.  
    '''
    __slots__ = ('_key', 'params', 'y', 'x', 'label', 'vel')

    def __init__(self, filebytes: bytes, data_info: FileBlockInfo, data_status_info: FileBlockInfo):
        self._key = data_info.get_data_key()
        self.params = Parameters(filebytes, [data_status_info])
        y = parse_data_block(filebytes, size=data_info.size, start=data_info.start, dpf=self.params.dpf)
        self.y = y[:self.params.npt]    # Trim extra values on some spectra
        self.x = np.linspace(self.params.fxv, self.params.lxv, self.params.npt)
        self.label = data_info.get_label()
        self.vel = 0

    def __getattr__(self, name):
        if name.lower() == 'wn' and self.params.dxu in ('WN', 'MI', 'LGW'):
            return self._get_wn()
        elif name.lower() == 'wl' and self.params.dxu in ('WN', 'MI', 'LGW'):
            return self._get_wl()
        elif name.lower() == 'f' and self.params.dxu in ('WN', 'MI', 'LGW'):
            return self._get_freq()
        elif name.lower() in self.params.keys():
            return getattr(self.params, name.lower())
        elif name == 'datetime':
            return self.params.datetime
        else:
            text = str(name) + ' is not a valid attribute for Data: ' + str(self._key)
            raise AttributeError(text)

    def _get_wn(self):
        if self.params.dxu == 'WN':
            return self.x
        elif self.params.dxu == 'MI':
            return 10000. / self.x
        elif self.params.dxu == 'LGW':
            return np.exp(self.x)

    def _get_wl(self):
        if self.params.dxu == 'WN':
            return 10000. / self.x
        elif self.params.dxu == 'MI':
            return self.x
        elif self.params.dxu == 'LGW':
            return 10000 / np.exp(self.x)
    
    def _get_freq(self):
        vel = 1000 * np.float(self.vel) / 7900  # cm/s
        return vel * self.wn


class Data3D(Data):
    '''Class containing 3D array data (series of spectra) and associated parameter/metadata from an OPUS file.

    Args:
        filebytes: raw bytes from OPUS file. see: read_opus_file_bytes
        data_info: FileBlockInfo instance of a 3D data block
        data_status_info: FileBlockInfo instance of a data status block which contains metadata about the data_info
            block. This block is a parameter block.

    Attributes:
        params: Parameter class with metadata associated with the data block such as first x point (fxp), last x point
            (lxp), number of points (npt), date (dat), time (tim) etc.
        y: 2D numpy array containing y values of data block
        x: 1D numpy array containing x values of data block. Units of x array are given by .dxu attribute.
        num_spectra: number of spectra in the series (i.e. length of y)
        label: human-readable string label describing the data block (e.g. Sample Spectrum, Absorbance, etc.)

    Extended Attributes:
        **wn:** Returns the x array in wavenumber (cm⁻¹) units regardless of what units the x array was originally saved
            in. This is only valid for spectral data blocks such as sample, reference, transmission, etc., not
            interferogram or phase blocks.  
        **wl:** Returns the x array in wavelength (µm) units regardless of what units the x array was originally saved
            in. This is only valid for spectral data blocks such as sample, reference, transmission, etc., not
            interferogram or phase blocks.  
        **datetime:** Returns a `datetime` class of when the data was taken (extracted from data status parameter
            block).  
        **xxx:** the various three char parameter keys from the "params" attribute can be directly called from the data
            class for convenience. Several of these parameters return arrays, rather than singular values because they
            are recorded for every spectra in the series, e.g. `npt`, `mny`, `mxy`, `tim`, `nsn`.  
    '''
    __slots__ = ('_key', 'params', 'y', 'x', 'num_spectra', 'label')

    def __init__(self, filebytes: bytes, data_info: FileBlockInfo, data_status_info: FileBlockInfo):
        self._key = data_info.get_data_key()
        self.params = Parameters(filebytes, [data_status_info])
        data = parse_3d_data_block(filebytes, start=data_info.start, dpf=self.params.dpf)
        self.y = data['y'][:, :self.params.npt]    # Trim extra values on some spectra
        self.x = np.linspace(self.params.fxv, self.params.lxv, self.params.npt)
        self.num_spectra = data['num_blocks']
        for key, val in data.items():
            if key not in ['y', 'version', 'offset', 'num_blocks', 'data_size', 'info_size']:
                self.params._params[key] = val
        self.label = data_info.get_label()


class Parameters:
    '''Class containing parameter metadata of an OPUS file.

    Parameters of an OPUS file are stored as key, val pairs, where the key is always three chars.  For example, the
    beamsplitter is stored in the "bms" attribute, source in "src" etc.  A list of known keys, with friendly label can
    be found in `brukeropus.file.constants.PARAM_LABELS`.  The keys in an OPUS file are not case sensitive, and stored
    in all CAPS (i.e. `BMS`, `SRC`, etc.) but this class uses lower case keys to follow python convention.  The class is
    initialized from a list of parameter `FileBlockInfo`.  The key, val items in blocks of the list are combined into
    one parameter class, so care must be taken not to pass blocks that will overwrite each others keys.  Analagous to a
    dict, the keys, values, and (key, val) can be iterated over using the functions: `keys()`, `values()`, and `items()`
    respectively.

    Args:
        filebytes: raw bytes from OPUS file. see: `brukeropus.file.parser.read_opus_file_bytes`
        param_blocks: list of `FileBlockInfo`; every block in the list should be classified as a parameter block.

    Attributes:
        xxx: parameter attributes are stored as three char keys. Which keys are generated depends on the list of
            `FileBlockInfo` that is used to initialize the class. If input list contains a single data status
            `FileBlockInfo`, attributes will include: `fxv`, `lxv`, `npt` (first x-val, last x-val, number of points),
            etc. Other blocks produce attributes such as: `bms`, `src`, `apt` (beamsplitter, source, aperture) etc. A
            full list of keys available in a given Parameters instance are given by the `keys()` method.
        datetime: if blocks contain the keys: `dat` (date) and `tim` (time), the `datetime` attribute of this class will
            be set to a python `datetime` object. Currently, only data status blocks are known to have these keys. If
            `dat` and `tim` are not present in the class, the `datetime` attribute will return `None`.
    '''
    __slots__ = ('_params', 'datetime')

    def __init__(self, filebytes: bytes, param_blocks: list):
        self._params = dict()
        for block_info in param_blocks:
            params = {key.lower(): val for key, val in parse_param_block(filebytes, block_info.size, block_info.start)}
            self._params.update(params)
            block_info.keys = tuple(params.keys())
        self._set_datetime()

    def __getattr__(self, name):
        if name.lower() in self._params.keys():
            return self._params[name.lower()]
        else:
            text = str(name) + ' not a valid attribute. For list of valid parameter keys, use: .keys()'
            raise AttributeError(text)

    def __getitem__(self, item):
        return self._params.__getitem__(item)

    def _set_datetime(self):
        if 'dat' in self.keys() and 'tim' in self.keys():
            try:
                date_str = self.dat
                time_str = self.tim
                dt_str = date_str + '-' + time_str[:time_str.index(' (')]
                try:
                    fmt = '%d/%m/%Y-%H:%M:%S.%f'
                    dt = datetime.datetime.strptime(dt_str, fmt)
                except:
                    try:
                        fmt = '%Y/%m/%d-%H:%M:%S.%f'
                        dt = datetime.datetime.strptime(dt_str, fmt)
                    except:
                        self.datetime = None
                self.datetime = dt
            except:
                self.datetime = None
        else:
            self.datetime = None

    def keys(self):
        '''Returns a `dict_keys` class of all valid keys in the class (i.e. dict.keys())'''
        return self._params.keys()

    def values(self):
        '''Returns a `dict_values` class of all the values in the class (i.e. dict.values())'''
        return self._params.values()

    def items(self):
        '''Returns a `dict_items` class of all the values in the class (i.e. dict.items())'''
        return self._params.items()


class FileDirectory:
    '''Contains type and pointer information for all blocks of data in an OPUS file.

    `FileDirectory` information is decoded from the raw file bytes of an OPUS file. First the header is read which
    provides the start location of the directory block, number of blocks in file, and maximum number of blocks the file
    supports. Then it decodes the block pointer information from each entry of the file's directory block. Rather than
    store all file blocks in a single list (as it is done in the OPUS file directory), this class sorts the blocks into
    categories: `data`, `data_status`, `params`, `rf_params`, `directory`, and `file_log`.  It also pairs the data
    blocks with their corresponding `data_status` block to simplify grouping y data with the parameters that are used to
    generate x data and other data block specific metadata.

    Args:
        filebytes: raw bytes from OPUS file. see: `brukeropus.file.parser.read_opus_file_bytes`

    Attributes:
        start: pointer to start location of the directory block
        max_blocks: maximum number of blocks supported by file
        num_blocks: total number of blocks in the file
        data_blocks: list of `FileBlockInfo` that contain array data (e.g. sample, reference, phase)
        data_status_blocks: list of `FileBlockInfo` that contain metadata specific to a data block (units, etc.)
        param_blocks: list of `FileBlockInfo` that contain metadata about the measurement sample
        rf_param_blocks: list of `FileBlockInfo` that contain metatdata about the reference measurement
        directory_block: `FileBlockInfo` for directory block that contains all the block info in the file
        file_log_block: `FileBlockInfo` of the file log (changes, etc.)
        data_and_status_block_pairs: (data: `FileBlockInfo`, data_status: `FileBlockInfo`) which pairs the data status
            parameter block (time, x units, y units, etc.) with the data block it informs
        unknown_blocks: list of `FileBlockInfo` with an unrecognized type (i.e. not sure how to parse)
    '''

    __slots__ = ('version', 'start', 'max_blocks', 'num_blocks', 'data_blocks', 'data_status_blocks', 'param_blocks',
                 'rf_param_blocks', 'directory_block', 'file_log_block', 'data_and_status_block_pairs',
                 'unknown_blocks')

    def __init__(self, filebytes: bytes):
        self.version, self.start, self.max_blocks, self.num_blocks = parse_header(filebytes)
        self.data_blocks: list = []
        self.data_status_blocks: list = []
        self.param_blocks: list = []
        self.rf_param_blocks: list = []
        self.directory_block: FileBlockInfo
        self.file_log_block: FileBlockInfo
        self.unknown_blocks: list = []
        for block_type, size, start in parse_directory(filebytes, self.start, self.num_blocks):
            block = FileBlockInfo(block_type=block_type, size=size, start=start)
            if block.is_data_status():
                self.data_status_blocks.append(block)
            elif block.is_rf_param():
                self.rf_param_blocks.append(block)
            elif block.is_param():
                self.param_blocks.append(block)
            elif block.is_directory():
                self.directory_block = block
            elif block.is_file_log():
                self.file_log_block = block
            elif block.is_data():
                self.data_blocks.append(block)
            else:
                block.bytes = filebytes[block.start:block.start + block.size]
                self.unknown_blocks.append(block)
        self.data_and_status_block_pairs = []
        self._pair_data_and_status_blocks()

    def __str__(self):
        data_keys = [b.get_data_key() for b in self.data_blocks]
        data_str = ', '.join(data_keys)
        return 'File Directory: ' + str(self.num_blocks) + ' total blocks; data blocks: (' + data_str + ')'

    def _pair_data_and_status_blocks(self):
        for data_block in self.data_blocks:
            status_matches = [block for block in self.data_status_blocks if block.is_data_status_match(data_block)]
            if len(status_matches) == 0:
                text = 'Warning: No data status block match for data block: ' + str(data_block) \
                    + '\n\tdata block will be ignored.'
                warnings.warn(text)
            elif len(status_matches) > 1:
                text = 'Warning: Multiple data status block matches for data block: ' + str(data_block) \
                    + '\n\tMatches:' + '; '.join([str(match) for match in status_matches]) \
                    + '\n\tdata block will be ignored.'
                warnings.warn(text)
            else:
                self.data_and_status_block_pairs.append((data_block, status_matches[0]))



import textwrap, re, os
from brukeropus.file.constants import TYPE_CODE_LABELS, PARAM_LABELS, CODE_3_ABR
from brukeropus.file.parser import (read_opus_file_bytes,
                                    parse_header,
                                    parse_directory,
                                    parse_data_block,
                                    parse_3d_data_block,
                                    parse_param_block,
                                    parse_file_log,
)


__all__ = ['find_opus_files', 'get_param_label', 'get_type_code_label', 'get_block_type_label', 'get_data_key',
           'parse_file_and_print']


__docformat__ = "google"


def find_opus_files(directory, recursive: bool = False):
    '''Finds all files in a directory with a strictly numeric extension (OPUS file convention).

    Returns a list of all files in directory that end in .# (e.g. file.0, file.1, file.1001, etc.). Setting recursive
    to true will search directory and all sub directories recursively. No attempt is made to verify the files are
    actually OPUS files (requires opening the file); the function simply looks for files that match the naming pattern.

    Args:
        directory (str or Path): path indicating directory to search
        recursive: Set to True to recursively search sub directories as well

    Returns:
        filepaths (list): list of filepaths that match OPUS naming convention (numeric extension)
    '''
    pattern = re.compile(r'.+\.[0-9]+$')
    file_list = []
    for root, dirnames, filenames in os.walk(directory):
        for filename in filenames:
            if pattern.match(filename):
                file_list.append(os.path.join(root, filename))
        if not recursive:
            break
    return file_list


def get_param_label(param: str):
    '''Returns a short but descriptive label for 3-letter parameters. For example, bms returns Beamsplitter.

    The 3-letter parameter input is not case sensitive.  This package includes the majority of parameters that OPUS
    uses, but in the event a parameter label is not known, this function will return: "Unknown XXX" where XXX is the
    unknown 3-letter parameter.

    Args:
        param: three letter parameter code (e.g. bms, src, npt, etc.) [not case sensitive]

    Returns:
        label (str): Human-readable string label for the parameter.
    '''
    try:
        return PARAM_LABELS[param.upper()]
    except KeyError:
        return 'Unknown ' + param.upper()


def get_type_code_label(pos_idx: int, val: int):
    '''Returns the type code label of a file block given the position index and value of the type code.

    The file blocks on an OPUS file feature six-integer type codes, for example (3, 1, 1, 2, 0, 0), that categorize the
    contents of the file block. The positional index defines the category, while the value at that index defines the
    specific type of that category.  For example, the first integer (pos_idx=0), describes the type of data in the
    block, if applicable:

        0: Undefined or N/A,
        1: Real Part of Complex Data,
        2: Imaginary Part of Complex Data,
        3: Amplitude

    This package includes the majority of type codes that OPUS uses, but in the event a type code label is not known,
    this function will return: "Unknown 0 4" where the first number is the position index, and the second is the
    unknown value integer.

    Args:
        pos_idx: positional index of the type code (0 - 5)
        val: value of the type code

    Returns:
        label (str): human-readable string label that describes the type code.
    '''
    try:
        return TYPE_CODE_LABELS[pos_idx][val]
    except KeyError:
        return 'Unknown ' + str(pos_idx) + ' ' + str(val)


def get_block_type_label(block_type: tuple):
    '''Converts a six-integer tuple block type into a human readable label.

    Args:
        block_type: six integer tuple found in the OPUS file directory that describes the block type

    Returns:
        label (str): human-readable string label
    '''
    labels = [get_type_code_label(idx, val) for idx, val in enumerate(block_type) if val > 0
              and get_type_code_label(idx, val) != '']
    return ' '.join(labels)


def get_data_key(block_type: tuple):
    '''Returns a shorthand key for a given data block type: sm, rf, igsm, a, t, r, etc.

    Determines if the data block type is an interferogram, single-channel, absorption, etc. and whether it is associated
    with the sample or reference channel and returns a shortand key-like label: sm, rf, igsm, igrf, a, t, r, etc.  For
    the full data label (e.g. Sample Spectrum, Absorbance) use: get_block_type_label.
    This package includes the majority of type codes that OPUS uses, but in the event a type code label is not known,
    this function will return: "_33" or "sm_33" where 33 will change to the unkown block_type integer value.

    Args:
        block_type: six integer tuple found in the OPUS file directory that describes the block type

    Returns:
        key (str): shorthand string label that can be utilized as a data key (e.g. "sm", "igrf", "a")'''
    if block_type[3] in CODE_3_ABR.keys():
        key = CODE_3_ABR[block_type[3]]
        if block_type[1] == 1:
            key = key + 'sm'
        elif block_type[1] == 2:
            key = key + 'rf'
        elif block_type[1] > 3:
            key = key + '_' + str(block_type[1])
    else:
        key = '_' + str(block_type[3])
        if block_type[1] == 1:
            key = 'sm' + key
        elif block_type[1] == 2:
            key = 'rf' + key
        elif block_type[1] > 3:
            key = '_' + str(block_type[1]) + key
    return key


def parse_file_and_print(filepath, width=120):
    '''Parses an OPUS file and prints the block information as it goes along to the console.

    This function demonstrates the basic usage and interaction of the parsing functions.  It
    can also be used to diagnose a file parsing issue if one comes up.

    Args:
        filepath (str or Path): filepath to an OPUS file.
    '''
    filebytes = read_opus_file_bytes(filepath)
    if filebytes is not None:
        width = 120
        info_col_widths = (28, 15, 16, 61)
        info_col_labels = ('Block Type', 'Size (bytes)', 'Start (bytes)', 'Friendly Name')
        _print_block_header(filepath, width)
        version, dir_start, max_blocks, num_blocks = parse_header(filebytes)
        h_text = '    '.join([
            'Version: ' + str(version),
            'Directory start: ' + str(dir_start),
            'Max Blocks: ' + str(max_blocks),
            'Num Blocks: ' + str(num_blocks)])
        _print_centered(h_text, width)
        _print_block_header('Directory', width)
        blocks = []
        _print_cols(info_col_labels, info_col_widths)
        for info in parse_directory(filebytes, dir_start, num_blocks):
            try:
                vals = [info[0], info[1], info[2], get_block_type_label(info[0])]
                _print_cols(vals, info_col_widths)
                blocks.append(info)
            except Exception as e:
                print('Exception parsing block info: ', e)
        for block_info in blocks:
            try:
                _parse_block_and_print(filebytes, block_info, width=width)
            except Exception as e:
                print('Exception parsing block:', block_info, '\n\tException:', e)
    else:
        print('Selected file is not an OPUS file: ', filepath)


def _print_block_header(label, width):
    'Helper function for: parse_file_and_print'
    print('\n' + '=' * width)
    _print_centered(label, width)


def _print_centered(text, width):
    'Helper function for: parse_file_and_print'
    print(' ' * int((width - len(text)) / 2) + text)


def _parse_block_and_print(filebytes, block_info, width):
    'Helper function for: parse_file_and_print'
    param_col_widths = (10, 45, 45)
    key_width = 10
    key_label_width = 45
    param_col_widths = (key_width, key_label_width, width - key_width - key_label_width)
    param_col_labels = ('Key', 'Friendly Name', 'Value')
    if block_info[0] != (0, 0, 0, 13, 0, 0):
        _print_block_header(get_block_type_label(block_info[0]), width)
        if block_info[0][2] > 0 or block_info[0] == (0, 0, 0, 0, 0, 1):
            _print_cols(param_col_labels, param_col_widths)
            for key, val in parse_param_block(filebytes, block_info[1], block_info[2]):
                _print_cols((key, get_param_label(key), val), param_col_widths)
        elif block_info[0] == (0, 0, 0, 0, 0, 5):
            log = parse_file_log(filebytes, block_info[1], block_info[2])
            for entry in log:
                for line in textwrap.wrap(entry, width=width):
                    print(line)
        elif block_info[0][5] == 2:
            data = parse_3d_data_block(filebytes, block_info[2])
            _print_centered('Num Blocks: ' + str(data['num_blocks']), width)
            _print_centered('Store Table: ' + str(data['store_table']), width)
            print(data['y'])
        elif block_info[0][0] > 0 and block_info[0][1] > 0 and block_info[0][2] == 0 and block_info[0][3] > 0:
            array = parse_data_block(filebytes, block_info[1], block_info[2])
            print(array)
        else:
            _print_centered('Undefined Block Type: Raw Bytes', width)
            print(filebytes[block_info[2]: block_info[2] + block_info[1]])


def _print_cols(vals, col_widths,):
    'Helper function for: parse_file_and_print'
    string = ''
    for i, val in enumerate(vals):
        col_width = col_widths[i]
        val = str(val)
        if len(val) <= col_width - 2:
            string = string + val + ' ' * (col_width - len(val))
        else:
            string = string + val[:col_width - 5] + '...  '
    print(string)
